%% Exact Non-Linear Smoother for DSGE Models
% by Michal Andrle (IMF)
%
% This tutorial demonstrates how to build your own tools 
% building on the flexibility and stable interface of IRIS.
% I am hugely indebted to Jaromir, without him I wouldn't work on this at
% all...
%
% An exact non-linear filter (smoother) for non-linear DSGE models is based
% on a simple idea of using non-linear least squares on top of repeated
% solution of the non-linear model. Kalman filter is a linear least
% squares, hence below you find a simple tool for non-linear least squares
% for IRIS models.
%
% The code of the filter |nlfilter.m| is suited mostly for smaller models
% with modest non-linearity, not for estimation. The limitation of this
% simple, truly a brute-force method is time and IRIS capability to solve
% the model every step of the way. Also, due to the solution method a
% Jensen inequality is ignored, a price paied for retainig a non-linearity.
%
% The tutorial clearly examplifies the importance and utility of IRIS'
% object oriented design with a well documented and stable public interface
% to its capabilities. The logic of the filter, however, will work with any
% other software able to solve a non-linear DSGE model.


%% Principles of the Exact Nonlinear Smoother
% The non-linear exact filter is based on a simple idea that if a linear
% Kalman filter is equivalent to linear least squares, a non-linear exact
% Kalman filter should can be based on non-linear least squares.
%
% This is simple and well understood in the literature. For purely
% backward-looking, recursive models this principle is exploited and used
% in the engineering literature. However, due to rational expectations the
% models in economics are more complex.
%
% The exact non-linear least squares filter is not feasible for estimation
% of the model, it would just take too much time. It is mainly for
% understanding of the shocks behind the non-linear model, identifed from
% the data if you have few minutes to spare. 
%
% Remember what we are doing here: we are solving a non-linear least squares problem
% on top of a model solution, searching for a path of shocks and set of
% initial conditions, that explain the observed data and are most likely
% given the likelihood of the shocks. The more periods and the more shocks,
% the larger is the problem.
%
% Let's dive into the example...



%% House keeping
clear all; close all; clear classes; clc


%% Loading the |Endocred| model
read_endocred;

%% Setting up the range and the model parameters
rng = qq(2000,1):qq(2012,4);

m1 = m;
m1.std_eps_y = 0.00001;
m1.std_eps_pi = 0.00001;
m1.std_ey = 0.03;
m1.std_epi = 0.07;


%% Get a random path of shocks to simulate a test case
% I resample the model in a linear way as a simple way of getting the
% shocks and paths of variables. Yet, as non-linear simulation as a base
% case |truth| is needed I re-simulate the model with nonlinearities.
% Allowing the |display| option visible in this case shows you the progress
% of the quasi Newton iteration of IRIS while solving a model... 

%d = sstatedb(m1,rng,'randomshocks=',true);
d = resample(m1,[],rng,1);

ds = simulate(m1,d,rng,'anticipate', false,...
                  'nonlinearise', length(rng),...
	      	  'maxiter', 5000,'display',false);

% clear screen			  
clc;			  

% just in case, let's remove the |credibility stock|
ds = rmfield(ds, 'c');

%% Comparing the Linear filter to Non-linear prediction error filter
% linear filter
[ans smth] = filter(m1,ds,rng);
dbf2 = smth.mean;

% filter with non-linear pred. error
[ans smth] = filter(m1,ds,rng,'nonlinearize',10);
dbf3 = smth.mean;

%% Exact non-linear smoother (brute force least squares)
% Now, let's compare also the non-linear least square approach to the
% linear filter and to the non-linear filter in IRIS, based on the
% non-linear prediction error.
%
% First, an instance of the class |nlfilter.m| needs to be initalized.
% Second, a solved IRIS model is passed to the filter, |initIrisModel()|.
% The user have full control on the way how the model is repeatedly
% simulated, i.e. maximum numbe of iteration, if shocks are surprises or if
% it is solved in a non-linear way at all. The standard model options are
% passed to |setSimulOpts| to initialize the filter. In an analogous way,
% the user can set the options for the least squares solver, see the help.
%
% Having the initialized filter, the filter is run using the |filter|
% method with the input being the IRIS database and range. Starting values
% are based on the steady-state of the model, though using |startDb| method
% you can provide your own, potentially...
%
% NOTE: This example will run for you if and only if you have access to
% Matlab optimizing Tbx. 
% For help and other options type |help/nlfilter| or investigate the code.


n = nlfilter();
  
  % initialize the |nlfilter|
  n.initIrisModel(m1);

  % option for a model simulation
  n.setSimulOpts('anticipate',false,...
	         'maxiter', 3000,...
                 'nonlinearize', length(rng),...
	         'display',false);
  
  % run the nl-filter  
  df1 = n.filter(ds,rng);


% credibility stock
f = figure();
	pp = plot([ds.c df1.c dbf2.c dbf3.c ]);
	  set(pp(1),'linewidth',4,'color','r');
          set(pp(2),'color','b','linewidth',1.5);
          set(pp(4),'color','k','linewidth',2,'linestyle','--');

	legend('TRUTH','nonlin filter','linear','nl-pred','location','best');
	title('credibility stock')

%% Conclusions
% As can be seen, the exact non-linar filter achieves more precision in
% this particular case. In principle it should be this way in each case.
% However, the price one pays for the increased precision, often just by a
% little, is not small -- ten minutes in this particular case on my super
% slow laptop. The sample
% of the simulation is realistic, but the size of the model and number of
% shocks is more on the smaller side than would be the case of models
% operated by central banks. In this case one would have to parallelize,
% which would speed up thing enormously as the problem is embarassignly
% parallel, or just run it over night....
%
% Kind regards
% MA



