%% NonLinear "kalman filter" using Large-Scale Least Squares Problem
%
% |nlfilter| class
%
%	methods:	filter(), initIrisModel(), setSimulOpts()
%			
%
%
% This is "brute force" algorithm, as it does not exploit the structure
% of the deterministic solution algorithm as it would be desirable. (Yet!)
%
%**********************************************************************
% Requires IRIS Tbx for Matlab version Feb. 2012 or newer
% (c) 2012 Michal Andrle, name_surname at yahoo ....
%
% First version: March 2012
% This version:  October 2012
%
% Many thanks to Jaromir Benes for ''soul support...'' ;)
%**********************************************************************
%
classdef nlfilter < handle
    
%% Comments & TODOs
% a] Consider both stationary & non-stationary models; stationary
%    first of course
% b] User supplied optimizers/nlsquares codes
% c] Allowing for parallelism (?) in computing the derivatives...
%


    

properties
    
    % Properties of the simulated model needed
    xnames = {};        % state variables names
    ynames = {};        % measurement variables names
    enames = {};        % names of the shocks
    exNames = {};
    eyNames = {};
    
    dbase  = [];        % IRIS tseries database
    dbss   = [];        % steady-staet database
    startdb = [];       % start database 
    range  = [];

    datavec = [];       % [NY*NPER x 1] vector of data
    exVariance = [];
    eyVariance = [];

    
    % Type of simulations to be used
    simul  = [];        % iris := 1, dynare := 0
    model  = [];
    ic     = [];        % initial conditions {names and lags}

    % State of the object
    init   = false;
    simcounter = 0;

    res    = [];
    xnow   = [];
    
    % options of the filtering
    options = {'option_simulate',{'anticipate',false,...
                                  'nonlinearize',Inf},...
              }; % TODO make default opts
    
    % options for optimisation
    optset = [];
    simopts = [];
    
end %-of properties block
    
    
    
       
    
methods    
    
    function self = nlfilter(varargin)
        self.init = false;
    end % of |constructor|
    
    function display(self) %--------------------------------------------------
    %% Standard display function
        fprintf('\t Exact Nonlinear Least-Square Filter Object.\n');
    end
    
    function self = initIrisModel(self,m) %--------------------------------------
    %% Initialize IRIS model for simulation and errors
    
        if ismodel(m) 
            % process IRIS model
            if ~issolved(m)
                try
                    m = solve(m); 
                catch
                    fprintf('@nlfilter:: Problem with IRIS model solution.\n');
                    fprintf('%s\n',lasterr);
                end               
            end

          % Simulation type := IRIS
            self.simul = 1; % each time IRIS will be called
            self.model = m;
                
          % Get names and use it for own purposes
            self.xnames = get(m,'xList');
            self.ynames = get(m,'yList');
            self.enames = get(m,'eList');
            self.exNames = get(m,'eXList');
            self.eyNames = get(m,'eYList');
         
          if isempty(self.ynames)
            error('@nlfilter:: Expected masurement variables... none found');
          end
         
          % Process initial conditions MAP
          self.ic = self.initconds();
         
          % Initialize the variances of the model 
          if isstationary(m)
            % use unconditional variance
            icList = self.ic.init_names;
            cc = acf(m,'order',0,'select',icList);
            self.ic.variance = diag(cc)+0.0001; % some |cc| can be zero due to log-linearized moments...
            self.ic.cc = cc;
            [u s v] = svd(cc); 
            self.ic.ccfactor = u*sqrt(s);
          else
            % use diffuse variance
            self.ic.variance = zeros(nic,1);
            self.ic.variance(:,1) = 100;
          end
        else
            error('@nlfilter:: Expected IRIS model object.\n'); 
        end
    
    end %end of |initIrisModel| -------------------------------------------%
    
    function self = initDynareModel(self,varargin)
    %% Initialize DYNARE model using global access to WORKSPACE
    % and |dynare_simulate.m| code a la GIMF/DYNARE
    %
    % XXX
    
    end % end of |initDynareModel| ---------------------------------------------%
    
    function  dout = filter(self,db, rng, varargin)
    %% This is the main function for processing the stuff    

        % assign a range
        self.range = rng;

        % Create a steady-state database as a bacground for simulation
        self.dbss = self.getssdb();

        % extract measured variables and store them in the filter obj.
        self.getData(db);

        % initialize variance of the model for use with |rng|
        self.initVariances();

        % store the database for a comparison
        self.dbase = db;

        % set up the optimizer
        myOptimset = self.getOptimset();
        
        % get initial conditions for the unknowns   
        [x0 xl xu] = self.getx0(db);

        % Call |lsqnonlin| and pass |self| so it can access |objfun|
        fprintf('---------------------------------------------------------------\n');
        fprintf('*** Starting the Exact NLS-Filter of the model. IN TESTING ***\n');
        fprintf('---------------------------------------------------------------\n');
        
        try 
          tic;
            [x,resnorm,residual,exitflag] = lsqnonlin(@objfun, x0, xl, xu, myOptimset, self);
          tt = toc;
        catch
          warning(lasterr);
        end

        fprintf('---------------------------------------------------------------\n');
        fprintf('*** Filter CPU time: %f seconds.\n',tt);
        fprintf('---------------------------------------------------------------\n');
 
        % turn the xpack |x| back into the database
        [res dout] = self.simulate(x);
        
    end %-of the filter --------------------------------------------------------%


   function [res] = objfun(x,self)
   %% This is the nonliner-least-squares filter (NLSF) objective function
     % use the xpack |x| and simulate the model to get residuals
     res = self.simulate(x);   
   end 

   function [opts] = getFilterOpts(self,useropts)
   %% proces user options
     self.options = passopt(self.options, useropts); 
   end

   function [x0 xl xu] = getx0(self,d)
   %% get initial conditions for a simulation based on a database input
     if isempty(self.startdb)
       x0 = self.db2xpack(self.dbss);
     else
       x0 = self.db2xpack(self.startdb);
     end
     [xl xu] = deal([]);
   end 

   function db = getssdb(self) %----------------------------------------------
   %% returns steady state database
     if self.simul == 1 % IRIS
       maxLag  = abs(get(self.model,'maxLag'));
       maxLead = abs(get(self.model,'maxLead'));
       ssrng = (self.range(1)-2*maxLag):(self.range(end)+maxLead);
   
       db = sstatedb(self.model,ssrng);
     else
       error('not ready yet');
     end
   end

   function db = startDb(self, d)
   %% provide initial conditions for a non-linear solver...
      self.startdb = d;
   end
    
   function  getData(self,db) %---------------------------------------------- 
   %% check if the database has all what we need
        missing = {};
        for j = 1:numel(self.ynames)
           if ~isfield(db,self.ynames{j})
               missing = [missing self.ynames{j}];
           end                     
        end
        if ~isempty(missing)
            msg = sprintf('@nlfilter:: following observables are missing from the database: %s\n',missing{:});
            error(msg);
        end

        % in case there are no errors, extract the data
        ymat = transpose(db2mat(db,self.range,self.ynames));
        self.datavec = ymat(:);
    end 

    
    function [res dout] = simulate(self,x) %---------------------------------------------- 
    %% Function to calculate the simulation of the model        
    % data and everything is available
    % |x| is a vector of initial conds and shocks
      if self.simul == 1
        [res dout] = self.simulateIris(x);
      else
        [res dout] = self.simulateDynare(x);
      end
    end

  
    function [res dsim] = simulateIris(self,x) %--------------------------------------
    %% Simulating IRIS model in a non-linear way for a vector        
    % simulates and vectorizes data, but DOES NOT rescale these data yet

      % check |x|
      self.xnow = x;
     
      % unpack |x| vector into the database
      d = xpack2db(self,x);
     
      % get the simulation options
      o = self.getSimulOpts();
     
      % simulate the model for given |x| specifying shocks and init conditions
      try
        dsim = simulate(self.model, d, self.range, o);
      catch
        warning(lasterr);
      end
     
      % check simulations internally
      self.simcounter = self.simcounter + 1;
     
      % pack the database into the |x| vector
      res = self.db2res(dsim);
      self.res = res;
     
      if isinf(res)
        error('The simulation is not progressing, Infs and NaNs occured...');
      end
    end

    function [db] = xpack2db(self,x)
    %% take the |x| pack and create a database needed, with IC and shocks
    %
    % |x| is the vector of 'parameters' to optimize for, which is
    %          a] transition shocks path
    %          b] initial conditions path

      % get dimensions
      ny = numel(self.eyNames);
      nx = numel(self.exNames);
      nic= numel(self.ic.init_names);
      nper = length(self.range);
  
      % database 
      db = sstatedb(self.model,self.range);
  
      % extract transition shocks and put to the database
      ex2reshape = x(1:nx*nper);
      ex_mat     = reshape(ex2reshape,nper,nx);
      for j = 1:numel(self.exNames)
        itsName = self.exNames{j};
        db.(itsName)(self.range) = ex_mat(:,j);
      end
  
      % extract initial conditions and put to the database
      aux = x(nx*nper+1:end);
      for j = 1:size(aux,1)
        itsName = self.ic.init_names{j};
        itsRng  = self.range(1)-self.ic.init_lags(j);
        db.(itsName)(itsRng) = aux(j);
      end

  % get the output back
  end %------------------------------------------------------------------------

  function x = db2xpack(self,d) %------------------------------------------------
  %% Takes the initial or any database |d| and creates xpack |x| having
  % transition shocks and initial conditions for the simulation
    
    % get dimensions
    ny = numel(self.eyNames);
    nx = numel(self.exNames);
    nic= numel(self.ic.init_names);
    nper = length(self.range);

    % pre-allocate the stuff (TODO: move a layer up to the object)
    x = zeros(nx*nper + nic,1);

    % put transition shocks there
    ex_ = transpose(db2mat(d,self.range,self.exNames));
    x(1:nx*nper,1) = ex_(:);

    % put initial conditions there
    aux = zeros(nic,1);
    for j = 1:nic
      itsName = self.ic.init_names{j};
      itsRng  = self.range(1)-self.ic.init_lags(j);
      aux(j) = d.(itsName)(itsRng); % this is value, not the |error|
    end
    x(nx*nper+1:end,1) = aux;

  end

    function res = db2res(self,d) %------------------------------------------------
    % calculates the RESIDUALS and init cond deviations
    % |res| = [transit_shocks' measurement shocks' init_conds']

    % get dimensions
    ny = numel(self.eyNames);
    nx = numel(self.exNames);
    nic= numel(self.ic.init_names);

    nper = length(self.range);

    % declare the space
    res = zeros(nx*nper + ny*nper + nic,1);

    % pack all transition shocks
    ex_ = transpose(db2mat(d,self.range,self.exNames));
    res(1:nx*nper) = (ex_(:))./self.exVariance; %-element-by-element division

    % calculate! & pack all measurement shocks
    ey_ = transpose(db2mat(d,self.range,self.ynames)); %-element-by-element division
    res_ey = ey_(:)-self.datavec;
    res(nx*nper+1:(nx+ny)*nper) = res_ey./self.eyVariance;

    % pack all the initial conditions
    % TODO: this is of course wrong, shortcut. needs to be ic*OmgIC*ic'
    aux = zeros(nic,1);
    for j = 1:nic
      itsName = self.ic.init_names{j};
      itsRng  = self.range(1)-self.ic.init_lags(j);
      itsVariance = self.ic.variance(j);
      %aux(j) = (d.(itsName)(itsRng) - self.ic.init_vals(j) );%/itsVariance; % this is value, not the |error|
      aux(j) = (d.(itsName)(itsRng) - self.ic.init_vals(j) )/itsVariance; % this is value, not the |error|
    end

    %put initial conditions error into the |res|
    res( (nx+ny)*nper+1 : end) = aux; % fill in the |x|
    %res( (nx+ny)*nper+1 : end) = transpose(aux'*self.ic.ccfactor); % fill in the |x|
    %res( (nx+ny)*nper+1 : end) = transpose(aux'*pinv(self.ic.cc)*aux); % fill in the |x|

    end % of the function
    

  function inits = initconds(self) %---------------------------------------------
  %% process initial conditions for the model

    % get the initial conditions list
    reqlist = get(self.model,'required');
    ttoks = regexp(reqlist,'[log\(]*(.*)[\)]*{(.*)}','tokens');

    % parse the list
    init_names = {}; 
    init_lags  = zeros(numel(reqlist),1);

    for j = 1:numel(reqlist)
      init_names{end+1} = ttoks{j}{1}{1};
      init_lags(j) = abs(str2double(char(ttoks{j}{1}(2))));
    end

    if isstationary(self.model)
      [flag errs listeq] = chksstate(self.model);
      if flag 
        sstate = get(self.model,'sstate');
        init_vals = zeros(numel(init_names),1);
        for j = 1:numel(init_names)
          init_vals(j) = real(sstate.(init_names{j})); 
        end
      end
    else
      error('This version has non-stationary models switched of... Oups ;)');
    end

    inits = struct;
    inits.init_names = init_names;
    inits.init_lags  = init_lags;
    inits.init_vals  = init_vals;

  end % get initial conditions 


  
  function self = initVariances(self)
  %% initializes variances for the |range| and the |model|
  % TODO: rewrite meaningfully, log-linearized variance could be
  %     misleading. for some varialbes it is zero by definiiton

    % get info
    nper = length(self.range);
    p    = get(self.model,'parameters');

    nex = numel(self.exNames);
    ney = numel(self.eyNames);
    
    % large vectors of variances
    self.exVariance = ones(nper*nex,1);
    self.eyVariance = ones(nper*ney,1);

    tmpx = zeros(nper,nex);
    % get variances of transition shocks
    for j=1:numel(self.exNames)
      pName = ['std_' self.exNames{j}];
      tmpx(:,j) = p.(pName);%^2;  
    end

    % get variances of measurement shocks
    tmpy = zeros(nper,ney);
    for j=1:numel(self.eyNames)
      pName = ['std_' self.eyNames{j}];     
      tmpy(:,j) = p.(pName)^2;  
    end

    self.exVariance = tmpx(:);
    self.eyVariance = tmpy(:);
  end % of the |initVariances|


  function o = setSimulOpts(self, varargin)
  %% set simulation options  
    self.simopts = cell2struct(varargin(2:2:end),varargin(1:2:end),2);
  end

  function o = getSimulOpts(self);
  %% returns simulation options
    if isempty(self.simopts)
      defopts = {'display',false,...
                 'anticipate',false,...
                 'nonlinearise', length(self.range),...
                 'maxiter',500,...
                };
        o = passopt(defopts,defopts);
    else  
        o = self.simopts;
    end
  end 

  function o = getOptimset(self)
  %% set up the optimization setup
    if isempty(self.optset)
      o = optimset('Display','iter',...
                   'Algorithm','trust-region-reflective',... 
                   'TolFun',1e-6,...
                   'TolX',1e-7,...
                   'MaxIter',24,...
                   'MaxFunEval',3000);
    else
      o = self.optset;    
    end
  end

  function self = setOptimset(self, optset)
  % set optimset
    self.optset = optset;
  end

    
end % of |methods| section    
    
    
end % of |nlfilter| class definition

